# import unittest
# import doctest
# import zizou.visualise

# def load_tests(tests):
#     tests.addTests(doctest.DocTestSuite(zizou.visualise))
#     return tests

# if __name__ == '__main__':
#     runner = unittest.TextTestRunner()
#     suite = unittest.TestSuite()
#     runner.run(load_tests(suite))

